# -*- coding: utf-8 -*-


u"""Модуль содержит класс, формирующий файл с информацией о сборке.

Класс BuildInfo при инстанцировании принимает парамет version_file_path,
содержащий директорию в которой будет храниться version.conf. При вызове
build.set_info() происходит определение параметров сборки и запись их
в файл version.conf. Если параметры сборки определить не удаётся, то
он пытается прочитать текущий version.conf пакета. Если он отсутствует,
то проставляются значение по-умолчанию. Класс используется в setup.py проекта.

Функция get_build_info принимает путь до файла version_file_path и
извлекает из него информацию о сборке. Эта информация, вместе с актуальным
окружением проекта возвращается в качестве ответа. Предполагается
использование функции для получения информации в окне "О системе".
"""

import os
import sys
import pip
import subprocess
from functools import partial
from datetime import datetime
from ConfigParser import ConfigParser, NoSectionError, NoOptionError


TEMPLATE = """
# Параметры этого файла управляют идентификацией версии проекта.
# Ручное изменение этих параметров приведет к невозможности
# нормальной установки обновлений.

[version]
BRANCH = {branch}
VERSION = {version}
REVISION = {revision}
VERSION_DATE = {version_date}
REVISION_DATE = {revision_date}
"""


class BuildInfo(object):

    u"""Предоставляет информацию о сборке.

    Информация извлекается из репозитария и окружения.
    Если же они не доступны, то из version.conf.

    Пример использования в setup.py:
    build = BuildInfo(/path/to/version_conf)
    build.set_info()

    setup(
        ...
        version=build.version
        ...
    )
    """

    def __init__(self, version_file_path):
        self._tag = self._branch = self._change_set = None
        self._version_file = os.path.join(version_file_path, "version.conf")

        # Текущее содержимое файла version.conf
        self._current_version_cnf_content = get_build_info(version_file_path)

    @staticmethod
    def _exec(cmd_line):

        u"""Метод производит выполнение консольной команды."""

        command = subprocess.Popen(
            cmd_line.split(),
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        out, err = command.communicate()
        if err:
            raise StandardError("Error occurred: {0}".format(err))

        return out and out.strip()

    @property
    def branch(self):

        u"""Возвращает название ветки с которой выполняется сборка."""

        if not self._branch:
            try:
                branch = self._exec("git name-rev --name-only HEAD")
            except StandardError:
                branch = self._current_version_cnf_content["branch"]
            if not branch:
                raise StandardError("Branch not found")
            self._branch = branch
        return self._branch

    @property
    def change_set(self):

        u"""Возвращает хэш коммита с которого выполняется сборка."""

        if not self._change_set:
            try:
                change_set = self._exec("git rev-parse HEAD")
            except StandardError:
                change_set = self._current_version_cnf_content["revision"]
            if not change_set:
                raise StandardError("ChangeSet not found")
            self._change_set = change_set

        return self._change_set

    @property
    def tag(self):

        u"""Возвращает тэг."""

        # метка не поставлена
        if not self._tag:
            try:
                tag = self._exec("git name-rev --tags --name-only HEAD")
            except StandardError:
                tag = "undefined"
            if not tag:
                raise StandardError("Tag not found!")
            # аннотированные тэги могут выводиться с доп. информацией
            # с символами ~ и ^
            tag = tag.split("^")[0].split("~")[0]
            self._tag = tag

        # метка поставлена
        return self._tag

    @property
    def version(self):

        u"""Возвращает версию сборки."""

        ver = self.tag

        if ver == 'undefined':
            # если делаем билд(т.е. тег не указан), то в версию
            # вместо тега ставим <название ветки.ревизия[:10]>
            ver = self.branch + '.' + self.change_set[:10]

        return ver

    def _get_content(self):

        u"""Формирование содержимого файла."""

        now = datetime.now().strftime("%d.%m.%Y")

        return TEMPLATE.format(
            branch=self.branch,
            version=self.version,
            revision=self.change_set,
            version_date=now,
            revision_date=now,
        )

    def set_info(self):

        u"""Определяет параметры сборки и записывает в файл."""

        content = self._get_content()
        with open(self._version_file, 'w') as file_version:
            file_version.write(content)
        sys.stdout.write(
            "\nVersion was set to '%s'\n" % self.version)


def get_build_info(version_conf_file_path):

    u"""Отдача информации о сборке.

    Функция возвращает информацию о сборке проекта.
    Может использоваться для получения информации в окне "О системе".
    :param version_conf_file_path: полный путь до файла с информацией о сборке.

    Пример использования:
    // setup.py
    build = BuildInfo(/path/to/version_conf)
    build.set_info()

    setup(
        ...
        version=build.version
        ...
    )
    // В проекте
    info = get_build_info(/path/to/version_conf)
    """
    parser = ConfigParser(defaults={"branch": "null"})
    version_conf_file = os.path.join(version_conf_file_path, "version.conf")
    parser.read(version_conf_file)
    parser_get = partial(parser.get, "version")

    try:
        data = {
            "version": parser_get("VERSION"),
            "revision": parser_get("REVISION"),
            "branch": parser_get("BRANCH"),
            "version_date": parser_get("VERSION_DATE"),
            "revision_date": parser_get("REVISION_DATE")
        }
    except (NoSectionError, NoOptionError):
        data = dict.fromkeys(
            [
                "version",
                "revision",
                "branch",
                "version_date",
                "revision_date"
            ], "-"
        )

    data["environment"] = pip.get_installed_distributions()

    return data
